# This file is part of CompressiveLearning.jl, a Julia package for 
# large-scale machine learning using sketches of generalized moments.
# Copyright (C) 2020 Antoine Chatalic
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ########################################################################

module CompressiveLearning

# Sketching & Learning
export SketchingOperator, BaseSkOp, CompositeSkOp, LinearSkOp, ComposedLinearSkOp
export MinMaxSkOp
export FourierSkOp, QuantizedFourierSkOp, TriangleFourierSkOp
export SqRandomSkOp
export FeaturesSubsamplingSkOp, SamplesSubsamplingSkOp
export base_skop, composed_skop, skops_pair # To generate sketching operators
export UDPAverageSkOp, AddOrRemoveOneSample, ReplaceOneSample, DP_skop	# privacy
export drawfrequencies, GaussianKernel, LaplacianKernel, AdaptedRadiusKernel
export sketch, sketch_and_bounds
export kmeans_inverse_problem, pca_inverse_problem, GMM_inverse_problem
export CLOMPR, CLAMP, CKM, CPCA, CGMM		# wrappers
export build_skop
# Datasets
export dataset_stats
# Evaluation
export NSSE, RSSE, evaluate_kmeans, kmeans_multitrials, ARI
export KL, symKL, evaluate_GMM
export evaluate_pca
# Fast Transforms
export FastTransform, WHOrthoRandom, Fastfood
export rademacher
export vcis, TRIGO_LIB
export test_ft

# const libfht = joinpath(dirname(@__DIR__), "deps", "libfasttransforms.so")
using FastHadamardStructuredTransforms_jll
const libfht = FastHadamardStructuredTransforms_jll.libfasttransforms

# using Base.Threads, LinearAlgebra, Random, Distributed, Requires, VML
using Base.Threads, LinearAlgebra, Random, Distributed, Requires
using Markdown			# for documentation
using StatsBase			# to sample without replacement
using Distributions
using Interpolations	# to sample from pdfs
using Clustering		# to compare with kmeans
using Suppressor 		# to silence LBFGSB output
using Distances
using BatchIterators, ProgressMeter 	# ProgressMeter for sketching
using SpecialFunctions, Roots	 # for Balle mechanism in privacy.jl (erf and find_root)
using DocStringExtensions # for SIGNATURES and other extensions
# Also requires LBFGS (cf. src/helpers/LBFGSB_wrapper.jl)
# Also requires some optimization libraries (cf src/learning_tasks/PCA.jl)

# Third party wrappers
include("helpers/LBFGSB_wrapper.jl")
include("helpers/helpers.jl")

# Code of the module
include("freqs.jl")	
include("sketching.jl")	
include("kernel_learning.jl")	
include("fasttransforms.jl")	
include("quantization.jl")	

# Applications
include("learning.jl")
include("learning_tasks/clustering.jl")
include("learning_tasks/GMM.jl")
include("learning_tasks/PCA.jl")

# Differential Privacy
include("privacy.jl")	

# Decoders
include("decoders/clompr.jl")		# CL-OMPR decoder
include("decoders/gamp.jl")			# CL-AMP decoder

function __init__()
	@require IntelVectorMath="c8ce9da6-5d36-5c03-b118-5a70151be7bc" begin
		using .VML
		include("helpers/vml.jl")
	end
end

end
