# This file is part of CompressiveLearning.jl, a Julia package for 
# large-scale machine learning using sketches of generalized moments.
# Copyright (C) 2020 Antoine Chatalic
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ########################################################################

# Related paper:
#
# [1] Sketching for large-scale learning of mixture models
# 	  Information and Inference: A Journal of the IMA, 7:3 (2017), pp. 447--508
# 	  N. Keriven, A. Bourrier, R. Gribonval, P. Pérez

using GaussianMixtures

"""
$(TYPEDEF)
Type for mixtures of Gaussians with diagonal covariances.
"""
mutable struct GaussianMixture <: MixtureType
	p::Int 		# Parameters dimension
	d::Int		# Input space dimension
end

#######################################################################
#                CL-OMPR internals for GaussianMixture                #
#######################################################################

"""
Returns a new atom.
"""
function init_param(θ, M::GaussianMixture; 
					init_method = :uniform, 
					σX = nothing,
					bounds=(-Inf,Inf), 
					kwargs...)

	@assert ~(σX == nothing)
	lb, ub = bounds
	bounds_μ = if (length(lb) > 1) 
		(lb[1:M.d], ub[1:M.d])
	elseif bounds end
	init_μ = init_param(θ[1:M.d,:], DiracsMixture(M.d), 
						init_method = init_method, 
						bounds = bounds_μ)
	init_σ² = σX^2 * (0.5 .+ rand()*ones(M.d));
	return [init_μ; init_σ²]
end

function sketchparams!(out, A::FourierSkOp, θ, M::GaussianMixture)
	Ω = A.Ω
	# Compute sketchs
	sketchcolumns!(out, A, θ[1:M.d,:])	
	out .*= exp.(-0.5 * (Ω.^2)' * θ[M.d+1:end,:])
	# Express the gradients
	function ∇skθ!(g, y)
		G = reshape((@view g[:]), M.p, :)	# 2d × k matrix corresponding to [∇μ; ∇Σ]
		# There is a sign error in Nicolas' notes for that:
		G[1:M.d, :] .= Ω*(real(out).*imag(y) - imag(out).*real(y))					# ∇μ
		G[M.d+1:end, :] .= -0.5 * Ω.^2 *(real(out).*real(y) + imag(out).*imag(y))	# ∇Σ
	end
	(out, ∇skθ!)
end

#######################################################################
#                              Wrappers                               #
#######################################################################

"""
	Compressive Gaussian Mixture Modeling
"""
function CGMM(X, k; m_factor = 10, 
			 		kernel_var = nothing,	# Chooses σ²=kernel_var
			 		kernel_var_inter_var_ratio = nothing, # Chooses σ²=kernel_var_inter_var_ratio*inter_var
			 		kernel_var_choice = nothing, 
					Fourier_renormalization = true,
					sketching_activation = :sinusoid, 
					learning_activation = :sinusoid,
			 		decoder = :CLOMPR,
			 		decoder_kwargs = Dict(),
			 		out_dic = nothing,
			 		kwargs...)

	radX = maximum(sqrt.(sum(X .^2, dims=1)))
	σX = sqrt(sum(sum(b.^2) for b in centered_batch_iterator(X))/(size(X,1)*size(X,2)))
	σ = choose_std(X; 
				   kernel_var = kernel_var, 
				   kernel_var_inter_var_ratio = kernel_var_inter_var_ratio, 
				   kernel_var_choice = kernel_var_choice, 
				   out_dic = out_dic)
	d = size(X, 1)
	m = Int(round(m_factor*k*d))		 # Number of frequencies
	@debug "CGMM with m=$(m) frequencies (d=$(d), k=$(k), m_factor=$(m_factor))."

	# Generate sketching operators
	(Ask, Aln) = skops_pair(m, d, σ; 
							sketching_activation = sketching_activation,
							learning_activation = learning_activation,
							out_dic = out_dic,
							kwargs...)

	# 1. Compute the sketch
	((s, bds), sketching_time) = @timed sketch_and_bounds(Ask, X)
	if Fourier_renormalization && (sketching_activation != learning_activation)
		s .*= quantization_renormalization(sketching_activation, learning_activation) end

 	# 2. Recover centroids and weights by solving the inverse problem
	((θ, α), learning_time) = @timed GMM_inverse_problem(Aln, s, k; 
										bounds = bds, 
										decoder = decoder,
										decoder_kwargs = decoder_kwargs,
										radX = radX,
										out_dic = out_dic)

	# If a dictionary out_dic is provided, store additional infos
	if !isa(out_dic, Nothing)
		out_dic[:m_factor] = m_factor
		out_dic[:m] = m
		out_dic[:decoder] = decoder
		out_dic[:sketching_time] = sketching_time
		out_dic[:learning_time] = learning_time
		out_dic[:Fourier_renormalization] = Fourier_renormalization
		for (k, v) in decoder_kwargs
			@assert typeof(k) <: Symbol
			out_dic[k] = decoder_kwargs[k]
		end
	end

	α ./= sum(α)
	GMM = MixtureModel(map(i -> MvNormal(θ[1:d, i], sqrt.(θ[d+1:end, i])), 1:k), α)
	return GMM
end

"""
$(SIGNATURES)
Returns a pair (θ,α) of GMM parameters and weights matching the sketch `s`.
Here θ is a d×k matrix whose columns are diracs, and α is a k-dimensional vector of weights.
"""
function GMM_inverse_problem(A, s, k ; 
								decoder = :CLOMPR, 
								bounds=(-Inf,Inf),
								decoder_kwargs = Dict(),
								min_σ² = 1e-5,
								radX = Inf,
								out_dic = Dict())
	p = 2A.d
	t = GaussianMixture(p, A.d)

	# Define bounds for parameters (stacking mean and diagonal covariance)
	lb, ub = bounds
	if (length(lb)==A.d) lb = [lb; min_σ²*ones(A.d)] end
	if (length(ub)==A.d) ub = [ub; radX*ones(A.d)] end
	if (lb isa Number) lb = repeat([lb, min_σ²], outer=A.d) end
	if (ub isa Number) ub = repeat([ub, radX], outer=A.d) end
	@assert length(lb) == length(ub) == p

	# Solve the inverse problem
	if decoder == :CLOMPR
		CLOMPR(s, k, t, A; 
			   bounds = (lb, ub),
			   out_dic = out_dic,
			   decoder_kwargs...)
	else
		error("Decoder $(algorithm) not supported for GMM fitting.")
	end
end

#######################################################################
#                             Evaluation                              #
#######################################################################

function KL(P, Q; n = 4e5, blocksize = 1000)
	kl = 0.0
	nblocks = ceil(Int, n/blocksize)
	for iblock = 1:nblocks
		XP = rand(P, blocksize)
		kl += sum(logpdf(P, XP) - logpdf(Q, XP))
	end
	kl/n
end
symKL(P, Q; kwargs...) = KL(P, Q; kwargs...) + KL(Q, P; kwargs...)

function evaluate_GMM(X, GMM, k, dic = Dict{Symbol, Any}(); 
					  compute_symKL = false,
					  compare_EM = false)
	dic[:loglhd] = sum(logpdf(GMM, X))
	if compute_symKL
		@assert haskey(dic, :todiscard) && haskey(dic[:todiscard], :dsGMM)
		trueGMM = dic[:todiscard][:dsGMM]
		dic[:symKL] = symKL(trueGMM, GMM)
	end
	if compare_EM
		GMM_EM = EM_GMM(X, k)
		dic[:loglhd_EM] = sum(logpdf(GMM_EM, X))
		if compute_symKL
			@assert haskey(dic, :todiscard) && haskey(dic[:todiscard], :dsGMM)
			trueGMM = dic[:todiscard][:dsGMM]
			dic[:symKL_EM] = symKL(trueGMM, GMM_EM)
			dic[:rsymKL] = dic[:symKL]/dic[:symKL_EM]
		end
	end
end

#######################################################################
#                            Other methods                            #
#######################################################################


function EM_GMM(X, k; its=10, out_dic=Dict(), kwargs...)
	Xt = copy(transpose(X))		# Sadly I don't see any way to avoid that given the API
	m = @suppress GaussianMixtures.GMM(k, Xt; 
									   method = :kmeans, 
									   kind = :diag, 
									   nInit = 20, 
									   nIter = its,
									   nFinal = its)
	MixtureModel(m) # TODO Not sure why it doesn't work anymore
end

