# This file is part of CompressiveLearning.jl, a Julia package for 
# large-scale machine learning using sketches of generalized moments.
# Copyright (C) 2020 Antoine Chatalic
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ########################################################################

using Test
using CompressiveLearning
using LinearAlgebra

k = 3
d = 8
n = 933
m = 33
X = randn(d, n)

@testset "CKM" begin
	C = CKM(X, k)	# Check that it runs without errors
	@test size(C) == (d, k)
end

@testset "Privacy" begin
	# Check that it runs without errors
	C = CKM(X, k; DP_ε = 1.0, DP_δ = 1e-4, DP_γ = 0.98)
	@test size(C) == (d, k)
	C = CKM(X, k; DP_ε = 1.0, DP_γ = 0.98)
	@test size(C) == (d, k)
end

@testset "FastTransforms" begin
	b = 3
	# Check that the dense representation of Ω' has expected row norms
	r = rand(b*d)
	Ω = WHOrthoRandom(d, b*d, r)
	denseΩ = Ω'*Matrix(1.0I, d, d)
	@test sqrt.(sum(denseΩ.^2, dims=2)) ≈ r 
end

@testset "Sketching Operators" begin
	for dist in [:Chi, :AdaptedRadius]
	for linop_type in [:dense, :HDHDHD, :HGHDHD, :fastfood]
		@test size(drawfrequencies(m, d, 1.0, 
								   linop_type = linop_type,
								   radialdist = dist)) == (d, m)
	end
	end
end

@testset "Sketching" begin
	d = 4
	n = 19
	m = 20
	X = randn(d, n)
	A = FourierSkOp(m, d, GaussianKernel())
	@testset "Batch Sketching" begin
		# Ensure we get the same results with different batch sizes
		s1 = sketch(A, X)
		s2 = sketch(A, X, 1)
		s3 = sketch(A, X, 3)
		s4 = sketch(A, X, 1298981223)
		@test s1 ≈ s2
		@test s1 ≈ s3
		@test s1 ≈ s4
	end
	@testset "Subsampling" begin
		# Test that the number of computed values per sample correspond to what is asked
		for i in [1, 3]
			Ass = FeaturesSubsamplingSkOp(A, i)
			X = randn(d, 1)
			s = sketch(Ass, X)
			@test sum(.~(s .≈ 0.0)) == i
		end
		# Test that with all observations, we get the same thing
		Ass = FeaturesSubsamplingSkOp(A, m)
		X = randn(d, n)
		@test all(sketch(Ass, X) .≈ sketch(A, X))
	end
end

@testset "Dimension 1" begin
	d = 1
	X = randn(d, n)
	C = CKM(X, k)	# Check that it runs without errors
	@test size(C) == (d, k)
end

